use crate::base::*;
use crate::parser::*;
use imbl::*;
use std::cell::RefCell;
use std::fmt;
use std::rc::Rc;

#[derive(Clone, Debug)]
pub struct Fn<'src> {
    pub name: &'src str,
    pub body: &'src Vec<MatchClause<'src>>,
}

#[derive(Debug)]
pub enum Value<'src> {
    Nil,
    Placeholder,
    Boolean(bool),
    Number(f64),
    Keyword(&'src str),
    String(&'src str),
    // on the heap for now
    Tuple(Rc<Vec<Self>>),
    Args(Rc<Vec<Self>>),
    // ref-counted, immutable, persistent
    List(Vector<Self>),
    // ref-counted, immutable, persistent
    Dict(HashMap<&'src str, Self>),
    Box(&'src str, Rc<RefCell<Self>>),
    Fn(Rc<Fn<'src>>),
    Base(Base<'src>),
    // Set(HashSet<Self>),
    // Sets are hard
    // Sets require Eq
    // Eq is not implemented on f64, because NaNs
    // We could use ordered_float::NotNan
    // Let's defer that
    // We're not really using sets in Ludus

    // Other things we're not implementing yet:
    // pkgs, nses, tests
}

impl<'src> Clone for Value<'src> {
    fn clone(&self) -> Value<'src> {
        match self {
            Value::Nil => Value::Nil,
            Value::Boolean(b) => Value::Boolean(b.clone()),
            Value::String(s) => Value::String(s),
            Value::Keyword(s) => Value::Keyword(s),
            Value::Number(n) => Value::Number(n.clone()),
            Value::Tuple(t) => Value::Tuple(t.clone()),
            Value::Args(a) => Value::Args(a.clone()),
            Value::Fn(f) => Value::Fn(f.clone()),
            Value::List(l) => Value::List(l.clone()),
            Value::Dict(d) => Value::Dict(d.clone()),
            Value::Box(name, b) => Value::Box(name, b.clone()),
            Value::Placeholder => Value::Placeholder,
            Value::Base(b) => Value::Base(b.clone()),
        }
    }
}

impl<'src> fmt::Display for Value<'src> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            Value::Nil => write!(f, "nil"),
            Value::Boolean(b) => write!(f, "{}", b),
            Value::Number(n) => write!(f, "{}", n),
            Value::Keyword(k) => write!(f, ":{}", k),
            Value::String(s) => write!(f, "\"{}\"", s),
            Value::Fn(fun) => write!(f, "fn {}", fun.name),
            Value::Tuple(t) | Value::Args(t) => write!(
                f,
                "({})",
                t.iter()
                    .map(|x| x.to_string())
                    .collect::<Vec<_>>()
                    .join(", ")
            ),
            Value::List(l) => write!(
                f,
                "[{}]",
                l.iter()
                    .map(|x| x.to_string())
                    .collect::<Vec<_>>()
                    .join(", ")
            ),
            Value::Dict(d) => write!(
                f,
                "#{{{}}}",
                d.iter()
                    .map(|(k, v)| format!(":{k} {v}"))
                    .collect::<Vec<_>>()
                    .join(", ")
            ),
            Value::Box(name, value) => {
                write!(
                    f,
                    "box {}: [{}]",
                    name,
                    &value.try_borrow().unwrap().to_string()
                )
            }
            Value::Placeholder => write!(f, "_"),
            Value::Base(base) => {
                let name = match base {
                    Base::Unary(name, _) => name,
                    Base::Binary(name, _) => name,
                    Base::Ternary(name, _) => name,
                };
                write!(f, "base fn {}", name)
            }
        }
    }
}

impl<'src> Value<'src> {
    pub fn bool(&self) -> bool {
        match self {
            Value::Nil | Value::Boolean(false) => false,
            _ => true,
        }
    }

    pub fn ludus_type(&self) -> Value {
        match self {
            Value::Nil => Value::Keyword("nil"),
            Value::Number(_) => Value::Keyword("number"),
            Value::Boolean(_) => Value::Keyword("boolean"),
            Value::Keyword(_) => Value::Keyword("keyword"),
            Value::Tuple(_) => Value::Keyword("tuple"),
            Value::String(_) => Value::Keyword("string"),
            Value::List(_) => Value::Keyword("list"),
            Value::Dict(_) => Value::Keyword("dict"),
            Value::Fn(_) => Value::Keyword("fn"),
            Value::Box(_, _) => Value::Keyword("box"),
            Value::Placeholder => unreachable!(),
            Value::Args(_) => unreachable!(),
            Value::Base(_) => Value::Keyword("fn"),
        }
    }
}

impl<'src> PartialEq for Value<'src> {
    fn eq(&self, other: &Value<'src>) -> bool {
        match (self, other) {
            // value equality types
            (Value::Nil, Value::Nil) => true,
            (Value::Boolean(x), Value::Boolean(y)) => x == y,
            (Value::Number(x), Value::Number(y)) => x == y,
            (Value::String(x), Value::String(y)) => x == y,
            (Value::Keyword(x), Value::Keyword(y)) => x == y,
            (Value::Tuple(x), Value::Tuple(y)) => x == y,
            (Value::List(x), Value::List(y)) => x == y,
            (Value::Dict(x), Value::Dict(y)) => x == y,
            // reference equality types
            (Value::Fn(x), Value::Fn(y)) => Rc::<Fn<'_>>::as_ptr(x) == Rc::<Fn<'_>>::as_ptr(y),
            (Value::Box(_, x), Value::Box(_, y)) => {
                Rc::<RefCell<Value<'_>>>::as_ptr(x) == Rc::<RefCell<Value<'_>>>::as_ptr(y)
            }
            _ => false,
        }
    }
}

impl<'src> Eq for Value<'src> {}
